@@ -2,203 +2,106 @@
 """
 Case insensitive derivable dictionary.
 
+Contents:
 
-Installation
-~~~~~~~~~~~~
-
-You can install the newest version of *pydicti* from PyPI:
-
-.. code:: bash
-
-    pip install pydicti
-
-Alternatively, you can just take the file ``pydicti.py`` and redistribute
-it with your application.
-
-
-Overview
-~~~~~~~~
-
-- ``class dicti``: default case insensitive dictionary type
-- ``class odicti``: ordered case insensitive dictionary type
-- ``def build_dicti``: create a case insensitive dictionary class
-- ``def Dicti``: create a case insensitive copy of a dictionary
-
-dicti
-=====
-
-Object of type ``dicti`` are dictionaries that feature case insensitive
-item access:
-
-.. code:: python
-
-    >>> d = dicti(Hello='foo', world='bar')
-    >>> d['heLLO']
-    'foo'
-    >>> 'WOrld' in d
-    True
-
-Internally however, the keys retain their original case:
-
-.. code:: python
-
-    >>> sorted(d.keys())
-    ['Hello', 'world']
-
-odicti
-======
-
-The type ``odicti`` instanciates order-preserving case insensitive
-dictionaries. It is available if either `collections.OrderedDict`_ or
-`ordereddict.OrderedDict`_ exists:
-
-.. code:: python
-
-    >>> odicti(zip('abc', range(3)))
-    odicti([('a', 0), ('b', 1), ('c', 2)])
-
-.. _`collections.OrderedDict`: http://docs.python.org/3.3/library/collections.html#collections.OrderedDict
-.. _`ordereddict.OrderedDict`: https://pypi.python.org/pypi/ordereddict/1.1
-
-build_dicti
-===========
-
-With ``build_dicti`` you can create custom case insensitive dictionaries.
-This function is what is used to create the ``pydicti.dicti`` and
-``pydicti.odicti`` types. Note that calling ``build_dicti`` several times
-with the same argument will result in identical types:
-
-.. code:: python
-
-    >>> build_dicti(dict) is dicti
-    True
-    >>> build_dicti(OrderedDict) is odicti
-    True
-
-``build_dicti`` uses subclassing to inherit the semantics of the given base
-dictionary type:
-
-.. code:: python
-
-    >>> issubclass(odicti, OrderedDict)
-    True
-
-Dicti
-=====
-
-The function ``Dicti`` is convenient for creating case insensitive
-copies of dictionary instances:
-
-.. code:: python
-
-    >>> o = OrderedDict(zip('abcdefg', range(7)))
-    >>> oi = Dicti(o)
-    >>> type(oi) is odicti
-    True
-
-
-JSON
-~~~~
-
-The subclassing approach allows to plug your dictionary instance into
-places where typechecking with ``isinstance`` is used, like in the json_
-module:
-
-.. code:: python
-
-    >>> import json
-    >>> d == json.loads(json.dumps(d), object_hook=dicti)
-    True
-
-.. _json: http://docs.python.org/3.3/library/json.html
-
-Above python26 you can use ``json.loads(s, object_pairs_hook=odicti)`` to
-deserialize ordered dictionaries.
-
-
-Pitfalls
-~~~~~~~~
-
-The equality comparison tries preserves the semantics of the base type as
-well as reflexitivity. This has impact on the transitivity of the
-comparison operator:
-
-.. code:: python
-
-    >>> i = dicti(oi)
-    >>> roi = odicti(reversed(list(oi.items())))
-    >>> roi == i and i == oi
-    True
-    >>> oi != roi and roi != oi  # NOT transitive!
-    True
-    >>> oi == i and i == oi      # reflexive
-    True
-
-The `coercion rules`_ in python allow this to work pretty well when
-performing comparisons between types that are subclasses of each other. Be
-careful otherwise, however.
-
-.. _`coercion rules`: http://docs.python.org/2/reference/datamodel.html#coercion-rules
-
-
-License
-~~~~~~~
-
-Copyright © 2013 Thomas Gläßle <t_glaessle@gmx.de>
-
-This work  is free. You can  redistribute it and/or modify  it under the
-terms of the Do What The Fuck  You Want To Public License, Version 2, as
-published by Sam Hocevar. See the COPYING file for more details.
-
-This program  is free software.  It comes  without any warranty,  to the
-extent permitted by applicable law.
-
+- class ``dicti``: case insensitive dictionary
+- class ``odicti``: case insensitive dictionary (ordered)
+- def ``build_dicti``: create a case insensitive dictionary class
+- def ``Dicti``: create a case insensitive copy of a given dictionary
 """
-__all__ = ['build_dicti', 'Dicti', 'odicti', 'dicti']
+
+__version__ = '1.1.5'
+
+__all__ = [
+    'dicti',
+    'odicti',
+    'Dicti',
+    'build_dicti',
+]
 
 import sys as _sys
-from collections import MutableMapping as _MutableMapping
-from copy import deepcopy as _deepcopy
-
 try:
-    from collections import OrderedDict
+    from collections.abc import MutableMapping as _MutableMapping
 except ImportError:
-    try:
-        from ordereddict import OrderedDict
-    except ImportError:
-        pass
+    from collections import MutableMapping as _MutableMapping
+from copy import deepcopy as _deepcopy
+from collections import OrderedDict
 
-# internally used to allow keys that are not strings
-def _lower(s):
-    """Convert to lower case if possible."""
+
+def normalize_case(s):
+    """
+    Convert to lower case if possible.
+
+    Used internally to allow keys that are not strings.
+    Will use casefold if available (Python >= 3.3).
+    """
     try:
-        return s.lower()
+        return s.casefold()
     except AttributeError:
-        return s
+        try:
+            return s.lower()
+        except AttributeError:
+            return s
+
 
 # make class
-def _make_dicti(dict_):
+def _make_dicti(dict_, normalize=normalize_case):
     _marker = []
+
     class Dicti(dict_):
         """
         Dictionary with case insensitive lookups.
 
         Behaves like  `dict` except  that lookups  are always  done case
-        insensitive. Case sensitivity  is retained in so  far that calls
-        like `.keys()` or  `.items()` return the keys  in their original
-        form.
+        insensitive, e.g.::
 
-        Passing  dictionaries   that  have  multiple  keys   with  equal
-        `.lower()`-case  to  the  constructor,   to  `.update()`  or  to
-        `.__eq__()` is undefined behaviour.
+            >>> d = dicti(Hello='foo', world='bar')
+            >>> d['heLLO']
+            'foo'
+            >>> 'WOrld' in d
+            True
+
+        However, case sensitivity  is retained in so  far that calls like
+        `.keys()` or `.items()` return the keys  in their original form::
+
+            >>> sorted(d.keys())
+            ['Hello', 'world']
+
+        This is compatible with json_ serialization::
+
+        .. code:: python
+
+            >>> import json
+            >>> d == json.loads(json.dumps(d), object_hook=dicti)
+            True
+
+        You can use ``json.loads(s, object_pairs_hook=odicti)`` to
+        deserialize ordered dictionaries.
 
         Note  that `dicti`  is  inherited  from `builtins.dict`  instead
         of   from   `collections.MutableMapping`.    This   means   that
         `isinstance(d,dict)` checks will  succeed and `json.dump()` will
         work automatically with `dicti`.
 
+        .. _json: http://docs.python.org/3.3/library/json.html
 
-        Implementation rationale (known pitfalls):
+
+        Ordered dictionaries
+        ====================
+
+        The type ``odicti`` instanciates order-preserving case insensitive
+        dictionaries::
+
+            >>> odicti(zip('abc', range(3)))
+            odicti({'a': 0, 'b': 1, 'c': 2})
+
+
+        Implementation rationale and known pitfalls
+        ===========================================
+
+        Passing  dictionaries   that  have  multiple  keys   with  equal
+        `.lower()`-case  to  the  constructor,   to  `.update()`  or  to
+        `.__eq__()` is undefined behaviour.
 
         When subclassing `builtins.dict`,  calling `dict(Dicti(v))` will
         bypass  the `__iter__`+`__getitem__`  interface  and access  the
@@ -211,34 +114,35 @@ def _make_dicti(dict_):
         problem. This easiest  way to do this is a  two step key lookup.
         The  internal dictionary  stores  (original_case  => value)  and
         another is used to store (lower_case => original_case).
-
         """
 
         # Constructor:
         def __init__(self, *args, **kwargs):
             """Initialize a case insensitive dictionary from the arguments."""
-            self.__case = {}        # create this member ASAP since
-                                    # dict_.__init__() might call self.clear()
+            # We have to setup `self.__case` at the very first, because
+            # dict_.__init__() might call self.clear():
+            self.__case = {}
             dict_.__init__(self)
             self.update(*args, **kwargs)
 
         # MutableMapping methods:
         def __getitem__(self, key):
             """Get the value for `key` case insensitively."""
-            return dict_.__getitem__(self, self.__case[_lower(key)])
+            return dict_.__getitem__(self, self.__case[normalize(key)])
 
         def __setitem__(self, key, value):
             """Set the value for `key` and assume new case."""
+            lower = normalize(key)
             # NOTE: this must be executed BEFORE dict_.__setitem__ in order
             # to leave a consistent state for base method:
-            if key in self:
-                del self[key]
+            if lower in self.__case:
+                key = self.__case[lower]
             dict_.__setitem__(self, key, value)
-            self.__case[_lower(key)] = key
+            self.__case[lower] = key
 
         def __delitem__(self, key):
             """Delete the item for `key` case insensitively."""
-            lower = _lower(key)
+            lower = normalize(key)
             dict_.__delitem__(self, self.__case[lower])
             # NOTE: this must be executed AFTER dict_.__delitem__ in order
             # to leave a consistent state for base method:
@@ -246,7 +150,7 @@ def _make_dicti(dict_):
 
         def __contains__(self, key):
             """Check if key is contained."""
-            return _lower(key) in self.__case
+            return normalize(key) in self.__case
 
         # Implemented by `dict_`
         # __iter__  # iterate in original case
@@ -256,10 +160,10 @@ def _make_dicti(dict_):
         # items / iteritems
 
         # Implemented by `MutableMapping`:
-        get          = _MutableMapping.get
-        popitem      = _MutableMapping.popitem
-        update       = _MutableMapping.update
-        setdefault   = _MutableMapping.setdefault
+        get        = _MutableMapping.get
+        popitem    = _MutableMapping.popitem
+        update     = _MutableMapping.update
+        setdefault = _MutableMapping.setdefault
 
         def clear(self):
             """Remove all entries from the dictionary."""
@@ -315,17 +219,19 @@ def _make_dicti(dict_):
         def __deepcopy__(self, memo):
             """Create a deep copy of the dictionary."""
             copy = self.__class__()
-            for k,v in self.items():
+            for k, v in self.items():
                 copy[k] = _deepcopy(v, memo)
             return copy
 
         def __repr__(self):
             """Representation string - something like `dicti([<items>])`."""
-            return '%s(%r)' % (self.__class__.__name__, list(self.items()))
+            return '%s(%s)' % (self.__class__.__name__, self)
 
         def __str__(self):
             """Display string - like the underlying dictionary."""
-            return '%r' % dict_(self.items())
+            return '{%s}' % ', '.join([
+                '%r: %r' % (k, v) for k, v in self.items()
+            ])
 
         # For now, let's assume that default pickling works fine for most
         # base classes. However, on python3 dict needs special treatment.
@@ -336,8 +242,10 @@ def _make_dicti(dict_):
         if dict_ is dict:
             def __reduce__(self):
                 return Dicti, (), self.__getstate__()
+
             def __getstate__(self):
                 return dict_(self)
+
             def __setstate__(self, state):
                 self.__case = {}
                 self.update(state.items())
@@ -345,7 +253,7 @@ def _make_dicti(dict_):
         # extra methods:
         def lower_items(self):
             """Iterate over (key,value) pairs with lowercase keys."""
-            return ((_lower(k), v) for k,v in self.items())
+            return ((normalize(k), v) for k, v in self.items())
 
         def lower_dict(self):
             """Return an underlying dictionary type with lowercase keys."""
@@ -356,54 +264,69 @@ def _make_dicti(dict_):
 
 
 _built_dicties = {}
-def build_dicti(base, name=None, module=None):
+
+
+def build_dicti(base, name=None, module=None, normalize=normalize_case):
     """
     Create a case insenstive subclass of `base`.
 
     :param MutableMapping base: base class
     :param str name: subclass name (defaults to base.__name__+'i')
     :param str module: module name for subclass (defaults to calling module)
+    :param callable normalize: normalization function that decides which keys
+                               map to the same values
 
     If  the class has already been created, this will not create a new type,
     but rather lookup the existing type in a table. The parameters `name`
     and `module` will not be used in this case.
 
+    Note that calling ``build_dicti`` several times with the same argument
+    will result in identical types::
+
+        >>> build_dicti(dict) is dicti
+        True
+        >>> build_dicti(OrderedDict) is odicti
+        True
+
+    ``build_dicti`` uses subclassing to inherit the semantics of the given
+    base dictionary type::
+
+        >>> issubclass(odicti, OrderedDict)
+        True
     """
     try:
-        cls = _built_dicties[base]
+        cls = _built_dicties[base, normalize]
     except KeyError:
         if not issubclass(base, _MutableMapping):
             raise TypeError("Not a mapping type: %s" % base)
-        cls = _make_dicti(base)
+        cls = _make_dicti(base, normalize)
         name = name or base.__name__ + 'i'
         cls.__name__ = name.rsplit('.', 1)[-1]
-        cls.__module__ = module or _sys._getframe(1).f_globals.get('__name__', '__main__')
+        cls.__module__ = module or _sys._getframe(1).f_globals.get(
+            '__name__', '__main__')
         cls.__qualname__ = name
-        _built_dicties[base] = cls
+        _built_dicties[base, normalize] = cls
     return cls
 
+
 def Dicti(obj):
     """
     Create case insensitive dictionary object from existing dictionary.
 
     The type of  `obj` is used as the type  of the underlying dictionary
-    for the returned case insensitive dictionary.
+    for the returned case insensitive dictionary::
+
+        >>> o = OrderedDict(zip('abcdefg', range(7)))
+        >>> oi = Dicti(o)
+        >>> type(oi) is odicti
+        True
 
     Since  this method  has the  same  name as  was used  to create  the
     class within  `build_dicti` this  allows `repr()`-esentations  to be
     invertible without further work.
-
     """
     return build_dicti(type(obj))(obj)
 
-# `dicti` is the default case insensitive dictionary
+
 dicti = build_dicti(dict)
-
-# `odicti` is an ordered, case insensitive dictionary type
-try:
-    OrderedDict
-except NameError:
-    pass
-else:
-    odicti = build_dicti(OrderedDict, 'odicti')
-
+odicti = build_dicti(OrderedDict, 'odicti')
